"""Dynamically define some metadata."""
import os

from hatchling.metadata.plugin.interface import MetadataHookInterface


def get_version_dev_status(root):
    """Get version_info without importing the entire module."""

    import importlib.util

    path = os.path.join(root, "soupsieve", "__meta__.py")
    spec = importlib.util.spec_from_file_location("__meta__", path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version_info__._get_dev_status()


class CustomMetadataHook(MetadataHookInterface):
    """Our metadata hook."""

    def update(self, metadata):
        """See https://ofek.dev/hatch/latest/plugins/metadata-hook/ for more information."""

        metadata["classifiers"] = [
            f"Development Status :: {get_version_dev_status(self.root)}",
            'Environment :: Console',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'Programming Language :: Python :: 3.12',
            'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Typing :: Typed'
        ]



"""Test empty selectors."""
from .. import util


class TestEmpty(util.TestCase):
    """Test empty selectors."""

    def test_empty(self):
        """Test empty."""

        markup = """
        <body>
        <div id="div">
        <p id="0" class="somewordshere">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3" class="herewords">Direct child</span>
        <pre id="pre" class="wordshere">
        <span id="4"> <!-- comment --> </span>
        <span id="5"> </span>
        <span id="6"></span>
        <span id="7"><span id="8"></span></span>
        </pre>
        </div>
        </body>
        """

        self.assert_selector(
            markup,
            "body :empty",
            ["4", "5", "6", "8"],
            flags=util.HTML
        )



"""Test first of type selectors."""
from .. import util


class TestFirstOfType(util.TestCase):
    """Test first of type selectors."""

    def test_first_of_type_at_start(self):
        """Test first of type which is also the first sibling."""

        markup = """
        <body>
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        </body>
        """

        self.assert_selector(
            markup,
            "p:first-of-type",
            ['0'],
            flags=util.HTML
        )

    def test_first_of_type_at_middle(self):
        """Test first of type that is not the first sibling."""

        markup = """
        <body>
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        </body>
        """

        self.assert_selector(
            markup,
            "span:first-of-type",
            ['2'],
            flags=util.HTML
        )

    def test_any_first_of_type(self):
        """Test any first of type."""

        markup = """
        <body>
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        </body>
        """
        self.assert_selector(
            markup,
            "body :first-of-type",
            ['0', '2'],
            flags=util.HTML
        )



"""Test contains selectors."""
from .. import util


class TestSoupContainsOwn(util.TestCase):
    """Test soup-contains-own selectors."""

    MARKUP = """
    <body>
    <div id="1">
    Testing
    <span id="2"> that </span>
    contains works.
    </div>
    </body>
    """

    def test_contains_own_descendants(self):
        """Test contains-own won't match text if contained in descendants."""

        self.assert_selector(
            self.MARKUP,
            'body div:-soup-contains-own(that)',
            [],
            flags=util.HTML
        )

    def test_contains_own(self):
        """Test contains-own."""

        self.assert_selector(
            self.MARKUP,
            'body *:-soup-contains-own(that)',
            ['2'],
            flags=util.HTML
        )

    def test_contains_own_cdata_html5(self):
        """Test contains CDATA in HTML5."""

        markup = """
        <body><div id="1">Testing that <span id="2"><![CDATA[that]]></span>contains works.</div></body>
        """

        self.assert_selector(
            markup,
            'body *:-soup-contains-own("that")',
            ['1'],
            flags=util.HTML5
        )

    def test_contains_own_cdata_py_html(self):
        """Test contains CDATA in Python HTML parser."""

        markup = """
        <body><div id="1">Testing that <span id="2"><![CDATA[that]]></span>contains works.</div></body>
        """

        self.assert_selector(
            markup,
            'body *:-soup-contains-own("that")',
            ['1'],
            flags=util.PYHTML
        )

    @util.skip_no_lxml
    def test_contains_own_cdata_lxml_html(self):
        """Test contains CDATA in `lxml` HTML."""

        from lxml import etree
        libxml_ver = etree.LIBXML_VERSION

        markup = """
        <body><div id="1">Testing that <span id="2"><![CDATA[that]]></span>contains works.</div></body>
        """

        results = ['1', '2'] if libxml_ver >= (2, 9, 11) else ['1']
        self.assert_selector(
            markup,
            'body *:-soup-contains-own("that")',
            results,
            flags=util.LXML_HTML
        )

    def test_contains_own_cdata_xml(self):
        """Test contains-own CDATA in XML."""

        markup = """
        <div id="1">Testing that <span id="2"><![CDATA[that]]></span>contains works.</div>
        """

        self.assert_selector(
            markup,
            '*:-soup-contains-own("that")',
            ['1', '2'],
            flags=util.XML
        )

    def test_contains_own_with_broken_text(self):
        """Test contains-own to see how it matches a broken text."""

        markup = """
        <body>
        <div id="1"> A simple test <div id="2"> to </div> show the broken text case. </div>
        </body>
        """
        self.assert_selector(
            markup,
            'body div:-soup-contains-own("test  show")',
            [],
            flags=util.HTML
        )



"""Test read write selectors."""
from .. import util


class TestReadWrite(util.TestCase):
    """Test read write selectors."""

    def test_read_write(self):
        """Test read write."""

        markup = """
        <input id="0">
        <textarea id="1"></textarea>

        <input id="2">
        <input id="3" disabled>

        <input id="4" type="email">
        <input id="5" type="number">
        <input id="6" type="password">
        <input id="7" type="search">
        <input id="8" type="tel">
        <input id="9" type="text">
        <input id="10" type="url">
        <input id="11" type="">
        <input id="12" type>

        <input id="13" type="button">
        <input id="14" type="checkbox">
        <input id="15" type="color">
        <input id="16" type="date">
        <input id="17" type="datetime-local">
        <input id="18" type="file">
        <input id="19" type="hidden">
        <input id="20" type="image">
        <input id="21" type="month">
        <input id="22" type="radio">
        <input id="23" type="range">
        <input id="24" type="reset">
        <input id="25" type="submit">
        <input id="26" type="time">
        <input id="27" type="week">

        <p id="28" contenteditable="">Text</p>
        <p id="29" contenteditable="true">Text</p>
        <p id="30" contenteditable="TRUE">Text</p>
        <p id="31" contenteditable="false">Text</p>
        <p id="32">Text</p>

        <input id="33" type="number" readonly>
        """

        self.assert_selector(
            markup,
            ":read-write",
            [
                '0', '1', '2', '4', '5', '6', '7', '8', '9', '10', '11',
                '12', '16', '17', '21', '26', '27', '28', '29', '30'
            ],
            flags=util.HTML
        )



"""Test ID selectors."""
from .. import util
from soupsieve import SelectorSyntaxError


class TestId(util.TestCase):
    """Test ID selectors."""

    MARKUP = """
    <div>
    <p>Some text <span id="1"> in a paragraph</span>.
    <a id="2" href="http://google.com">Link</a>
    </p>
    </div>
    """

    def test_id(self):
        """Test ID."""

        self.assert_selector(
            self.MARKUP,
            "#\\31",
            ["1"],
            flags=util.HTML
        )

    def test_tag_and_id(self):
        """Test tag and ID."""

        self.assert_selector(
            self.MARKUP,
            "a#\\32",
            ["2"],
            flags=util.HTML
        )

    def test_malformed_id(self):
        """Test malformed ID."""

        # Malformed id
        self.assert_raises('td#.some-class', SelectorSyntaxError)



"""Test `nth` last child selectors."""
from .. import util


class TestNthLastChild(util.TestCase):
    """Test `nth` last child selectors."""

    def test_nth_last_child(self):
        """Test `nth` last child."""

        markup = """
        <body>
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        </body>
        """

        self.assert_selector(
            markup,
            "p:nth-last-child(2)",
            ['10'],
            flags=util.HTML
        )

    def test_nth_last_child_complex(self):
        """Test `nth` last child complex."""

        markup = """
        <body>
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        </body>
        """

        self.assert_selector(
            markup,
            "p:nth-last-child(2n + 1)",
            ['1', '7', '9'],
            flags=util.HTML
        )



"""Test general pseudo-class cases."""
from .. import util


class TestPseudoClass(util.TestCase):
    """Test pseudo-classes."""

    def test_pseudo_class_not_implemented(self):
        """Test pseudo-class that is not implemented."""

        self.assert_raises(':not-implemented', NotImplementedError)

    def test_unrecognized_pseudo(self):
        """Test unrecognized pseudo class."""

        self.assert_raises(':before', NotImplementedError)



"""Test current selectors."""
from .. import util


class TestCurrent(util.TestCase):
    """Test current selectors."""

    MARKUP = """
    <body>
    <div id="div">
    <p id="0">Some text <span id="1" class="foo:bar:foobar"> in a paragraph</span>.
    <a id="2" class="bar" href="http://google.com">Link</a>
    <a id="3">Placeholder text.</a>
    </p>
    </div>
    </body>
    """

    def test_current(self):
        """Test current (should match nothing)."""

        self.assert_selector(
            self.MARKUP,
            "p:current",
            [],
            flags=util.HTML
        )

    def test_not_current(self):
        """Test not current."""

        self.assert_selector(
            self.MARKUP,
            "p:not(:current)",
            ["0"],
            flags=util.HTML
        )

    def test_current_func(self):
        """Test the functional form of current (should match nothing)."""

        self.assert_selector(
            self.MARKUP,
            ":current(p, div, a)",
            [],
            flags=util.HTML
        )

    def test_current_func_nested(self):
        """Test the nested functional form of current (should match nothing)."""

        self.assert_selector(
            self.MARKUP,
            ":current(p, :not(div), a)",
            [],
            flags=util.HTML
        )

        self.assert_selector(
            self.MARKUP,
            "body :not(:current(p, div, a))",
            ["div", "0", "1", "2", "3"],
            flags=util.HTML
        )

        self.assert_selector(
            self.MARKUP,
            "body :not(:current(p, :not(div), a))",
            ["div", "0", "1", "2", "3"],
            flags=util.HTML
        )



"""Test not selectors."""
from .. import util
from bs4 import BeautifulSoup
from soupsieve import SelectorSyntaxError


class TestNot(util.TestCase):
    """Test not selectors."""

    MARKUP = """
    <div>
    <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
    <a id="2" href="http://google.com">Link</a>
    <span id="3">Direct child</span>
    <pre id="pre">
    <span id="4">Child 1</span>
    <span id="5">Child 2</span>
    <span id="6">Child 3</span>
    </pre>
    </div>
    """

    def test_not(self):
        """Test not."""

        self.assert_selector(
            self.MARKUP,
            'div :not([id="1"])',
            ["0", "2", "3", "4", "5", "6", "pre"],
            flags=util.HTML
        )

    def test_not_and_type(self):
        """Test not with a tag."""

        self.assert_selector(
            self.MARKUP,
            'span:not([id="1"])',
            ["3", "4", "5", "6"],
            flags=util.HTML
        )

    def test_not_case(self):
        """Test not token case insensitivity."""

        self.assert_selector(
            self.MARKUP,
            'div :NOT([id="1"])',
            ["0", "2", "3", "4", "5", "6", "pre"],
            flags=util.HTML
        )

    def test_none_inputs(self):
        """Test weird inputs."""

        soup = BeautifulSoup('<span foo="something">text</span>', 'html.parser')
        soup.span['foo'] = None
        self.assertEqual(len(soup.select('span:not([foo])')), 0)

    def test_invalid_pseudo_empty(self):
        """Test pseudo class group with empty set."""

        self.assert_raises(':not()', SelectorSyntaxError)

    def test_invalid_pseudo_trailing_comma(self):
        """Test pseudo class group with trailing comma."""

        self.assert_raises(':not(.class,)', SelectorSyntaxError)

    def test_invalid_pseudo_leading_comma(self):
        """Test pseudo class group with leading comma."""

        self.assert_raises(':not(,.class)', SelectorSyntaxError)

    def test_invalid_pseudo_multi_comma(self):
        """Test pseudo class group with multiple commas."""

        self.assert_raises(':not(.this,,.that)', SelectorSyntaxError)



"""Test descendant combinators."""
from .. import util


class TestDescendants(util.TestCase):
    """Test descendant combinators."""

    def test_descendants(self):
        """Test descendants."""

        self.assert_selector(
            """
            <div>
            <p>Some text <span id="1"> in a paragraph</span>.
            <a id="2" href="http://google.com">Link</a>
            </p>
            </div>
            """,
            "div span",
            ["1"],
            flags=util.HTML
        )



"""Test matches selectors."""
from .. import util


class TestMatches(util.TestCase):
    """Test matches selectors."""

    MARKUP = """
    <div>
    <p>Some text <span id="1"> in a paragraph</span>.
    <a id="2" href="http://google.com">Link</a>
    </p>
    </div>
    """

    def test_matches(self):
        """Test multiple selectors with "matches"."""

        self.assert_selector(
            self.MARKUP,
            ":matches(span, a)",
            ["1", "2"],
            flags=util.HTML
        )

    def test_nested_matches(self):
        """Test multiple nested selectors with "matches"."""

        self.assert_selector(
            self.MARKUP,
            ":matches(span, a:matches(#\\32))",
            ["1", "2"],
            flags=util.HTML
        )



"""Test escapes."""
from .. import util


class TestEscapes(util.TestCase):
    """Test escapes."""

    def test_escapes(self):
        """Test escapes."""

        markup = """
        <div>
        <p>Some text <span id="1" class="foo:bar:foobar"> in a paragraph</span>.
        <a id="2" class="bar" href="http://google.com">Link</a>
        </p>
        </div>
        """

        self.assert_selector(
            markup,
            ".foo\\:bar\\3a foobar",
            ["1"],
            flags=util.HTML
        )



"""Test local link selectors."""
from .. import util


class TestLocalLink(util.TestCase):
    """Test local link selectors."""

    MARKUP = """
    <a id="1" href="./somelink/index.html">Link</link>
    <a id="2" href="http://somelink.com/somelink/index.html">Another link</a>
    """

    def test_local_link(self):
        """Test local link (matches nothing)."""

        self.assert_selector(
            self.MARKUP,
            "a:local-link",
            [],
            flags=util.HTML
        )

    def test_not_local_link(self):
        """Test not local link."""

        self.assert_selector(
            self.MARKUP,
            "a:not(:local-link)",
            ["1", "2"],
            flags=util.HTML
        )
